@@ -1,6 +1,8 @@
 from __future__ import print_function, division, absolute_import
 from __future__ import unicode_literals
+from fontTools.misc.py23 import *
 from fontTools.feaLib.error import FeatureLibError
+from fontTools.misc.encodingTools import getEncoding
 from collections import OrderedDict
 import itertools
 
@@ -45,31 +47,41 @@ def asFea(g):
         return g
 
 
-class Statement(object):
-    def __init__(self, location):
+class Element(object):
+
+    def __init__(self, location=None):
         self.location = location
 
     def build(self, builder):
         pass
 
     def asFea(self, indent=""):
-        pass
+        raise NotImplementedError
+
+    def __str__(self):
+        return self.asFea()
 
 
-class Expression(object):
-    def __init__(self, location):
-        self.location = location
+class Statement(Element):
+    pass
 
-    def build(self, builder):
-        pass
+
+class Expression(Element):
+    pass
+
+
+class Comment(Element):
+    def __init__(self, text, location=None):
+        super(Comment, self).__init__(location)
+        self.text = text
 
     def asFea(self, indent=""):
-        pass
+        return self.text
 
 
 class GlyphName(Expression):
     """A single glyph name, such as cedilla."""
-    def __init__(self, location, glyph):
+    def __init__(self, glyph, location=None):
         Expression.__init__(self, location)
         self.glyph = glyph
 
@@ -82,7 +94,7 @@ class GlyphName(Expression):
 
 class GlyphClass(Expression):
     """A glyph class, such as [acute cedilla grave]."""
-    def __init__(self, location, glyphs=None):
+    def __init__(self, glyphs=None, location=None):
         Expression.__init__(self, location)
         self.glyphs = glyphs if glyphs is not None else []
         self.original = []
@@ -130,7 +142,7 @@ class GlyphClass(Expression):
 
 class GlyphClassName(Expression):
     """A glyph class name, such as @FRENCH_MARKS."""
-    def __init__(self, location, glyphclass):
+    def __init__(self, glyphclass, location=None):
         Expression.__init__(self, location)
         assert isinstance(glyphclass, GlyphClassDefinition)
         self.glyphclass = glyphclass
@@ -144,7 +156,7 @@ class GlyphClassName(Expression):
 
 class MarkClassName(Expression):
     """A mark class name, such as @FRENCH_MARKS defined with markClass."""
-    def __init__(self, location, markClass):
+    def __init__(self, markClass, location=None):
         Expression.__init__(self, location)
         assert isinstance(markClass, MarkClass)
         self.markClass = markClass
@@ -157,7 +169,7 @@ class MarkClassName(Expression):
 
 
 class AnonymousBlock(Statement):
-    def __init__(self, tag, content, location):
+    def __init__(self, tag, content, location=None):
         Statement.__init__(self, location)
         self.tag, self.content = tag, content
 
@@ -169,7 +181,7 @@ class AnonymousBlock(Statement):
 
 
 class Block(Statement):
-    def __init__(self, location):
+    def __init__(self, location=None):
         Statement.__init__(self, location)
         self.statements = []
 
@@ -193,7 +205,7 @@ class FeatureFile(Block):
 
 
 class FeatureBlock(Block):
-    def __init__(self, location, name, use_extension):
+    def __init__(self, name, use_extension=False, location=None):
         Block.__init__(self, location)
         self.name, self.use_extension = name, use_extension
 
@@ -211,30 +223,32 @@ class FeatureBlock(Block):
         builder.end_feature()
 
     def asFea(self, indent=""):
-        res = indent + "feature {} {{\n".format(self.name.strip())
-        indent += SHIFT
-        if len(self.statements) and isinstance(self.statements[0], FeatureNameStatement):
-            res += indent + "featureNames {\n"
-            res += indent + SHIFT
-            res += ("\n" + indent + SHIFT).join(
-                [s.asFea(indent=indent + SHIFT * 2)
-                 for s in self.statements if isinstance(s, FeatureNameStatement)])
-            res += "\n"
-            res += indent + "};\n" + indent
-            res += ("\n" + indent).join(
-                [s.asFea(indent=indent)
-                 for s in self.statements if not isinstance(s, FeatureNameStatement)])
-            res += "\n"
-        else:
-            res += indent
-            res += ("\n" + indent).join([s.asFea(indent=indent) for s in self.statements])
-            res += "\n"
-        res += "{}}} {};\n".format(indent[:-len(SHIFT)], self.name.strip())
+        res = indent + "feature %s {\n" % self.name.strip()
+        res += Block.asFea(self, indent=indent)
+        res += indent + "} %s;\n" % self.name.strip()
+        return res
+
+
+class NestedBlock(Block):
+    def __init__(self, tag, block_name, location=None):
+        Block.__init__(self, location)
+        self.tag = tag
+        self.block_name = block_name
+
+    def build(self, builder):
+        Block.build(self, builder)
+        if self.block_name == "ParamUILabelNameID":
+            builder.add_to_cv_num_named_params(self.tag)
+
+    def asFea(self, indent=""):
+        res = "{}{} {{\n".format(indent, self.block_name)
+        res += Block.asFea(self, indent=indent)
+        res += "{}}};\n".format(indent)
         return res
 
 
 class LookupBlock(Block):
-    def __init__(self, location, name, use_extension):
+    def __init__(self, name, use_extension=False, location=None):
         Block.__init__(self, location)
         self.name, self.use_extension = name, use_extension
 
@@ -252,7 +266,7 @@ class LookupBlock(Block):
 
 
 class TableBlock(Block):
-    def __init__(self, location, name):
+    def __init__(self, name, location=None):
         Block.__init__(self, location)
         self.name = name
 
@@ -265,7 +279,7 @@ class TableBlock(Block):
 
 class GlyphClassDefinition(Statement):
     """Example: @UPPERCASE = [A-Z];"""
-    def __init__(self, location, name, glyphs):
+    def __init__(self, name, glyphs, location=None):
         Statement.__init__(self, location)
         self.name = name
         self.glyphs = glyphs
@@ -279,8 +293,8 @@ class GlyphClassDefinition(Statement):
 
 class GlyphClassDefStatement(Statement):
     """Example: GlyphClassDef @UPPERCASE, [B], [C], [D];"""
-    def __init__(self, location, baseGlyphs, markGlyphs,
-                 ligatureGlyphs, componentGlyphs):
+    def __init__(self, baseGlyphs, markGlyphs, ligatureGlyphs,
+                 componentGlyphs, location=None):
         Statement.__init__(self, location)
         self.baseGlyphs, self.markGlyphs = (baseGlyphs, markGlyphs)
         self.ligatureGlyphs = ligatureGlyphs
@@ -321,9 +335,13 @@ class MarkClass(object):
         for glyph in definition.glyphSet():
             if glyph in self.glyphs:
                 otherLoc = self.glyphs[glyph].location
+                if otherLoc is None:
+                    end = ""
+                else:
+                    end = " at %s:%d:%d" % (
+                        otherLoc[0], otherLoc[1], otherLoc[2])
                 raise FeatureLibError(
-                    "Glyph %s already defined at %s:%d:%d" % (
-                        glyph, otherLoc[0], otherLoc[1], otherLoc[2]),
+                    "Glyph %s already defined%s" % (glyph, end),
                     definition.location)
             self.glyphs[glyph] = definition
 
@@ -336,7 +354,7 @@ class MarkClass(object):
 
 
 class MarkClassDefinition(Statement):
-    def __init__(self, location, markClass, anchor, glyphs):
+    def __init__(self, markClass, anchor, glyphs, location=None):
         Statement.__init__(self, location)
         assert isinstance(markClass, MarkClass)
         assert isinstance(anchor, Anchor) and isinstance(glyphs, Expression)
@@ -352,7 +370,7 @@ class MarkClassDefinition(Statement):
 
 
 class AlternateSubstStatement(Statement):
-    def __init__(self, location, prefix, glyph, suffix, replacement):
+    def __init__(self, prefix, glyph, suffix, replacement, location=None):
         Statement.__init__(self, location)
         self.prefix, self.glyph, self.suffix = (prefix, glyph, suffix)
         self.replacement = replacement
@@ -384,8 +402,8 @@ class AlternateSubstStatement(Statement):
 
 
 class Anchor(Expression):
-    def __init__(self, location, name, x, y, contourpoint,
-                 xDeviceTable, yDeviceTable):
+    def __init__(self, x, y, name=None, contourpoint=None,
+                 xDeviceTable=None, yDeviceTable=None, location=None):
         Expression.__init__(self, location)
         self.name = name
         self.x, self.y, self.contourpoint = x, y, contourpoint
@@ -407,7 +425,7 @@ class Anchor(Expression):
 
 
 class AnchorDefinition(Statement):
-    def __init__(self, location, name, x, y, contourpoint):
+    def __init__(self, name, x, y, contourpoint=None, location=None):
         Statement.__init__(self, location)
         self.name, self.x, self.y, self.contourpoint = name, x, y, contourpoint
 
@@ -420,7 +438,7 @@ class AnchorDefinition(Statement):
 
 
 class AttachStatement(Statement):
-    def __init__(self, location, glyphs, contourPoints):
+    def __init__(self, glyphs, contourPoints, location=None):
         Statement.__init__(self, location)
         self.glyphs, self.contourPoints = (glyphs, contourPoints)
 
@@ -434,7 +452,7 @@ class AttachStatement(Statement):
 
 
 class ChainContextPosStatement(Statement):
-    def __init__(self, location, prefix, glyphs, suffix, lookups):
+    def __init__(self, prefix, glyphs, suffix, lookups, location=None):
         Statement.__init__(self, location)
         self.prefix, self.glyphs, self.suffix = prefix, glyphs, suffix
         self.lookups = lookups
@@ -466,7 +484,7 @@ class ChainContextPosStatement(Statement):
 
 
 class ChainContextSubstStatement(Statement):
-    def __init__(self, location, prefix, glyphs, suffix, lookups):
+    def __init__(self, prefix, glyphs, suffix, lookups, location=None):
         Statement.__init__(self, location)
         self.prefix, self.glyphs, self.suffix = prefix, glyphs, suffix
         self.lookups = lookups
@@ -498,7 +516,7 @@ class ChainContextSubstStatement(Statement):
 
 
 class CursivePosStatement(Statement):
-    def __init__(self, location, glyphclass, entryAnchor, exitAnchor):
+    def __init__(self, glyphclass, entryAnchor, exitAnchor, location=None):
         Statement.__init__(self, location)
         self.glyphclass = glyphclass
         self.entryAnchor, self.exitAnchor = entryAnchor, exitAnchor
@@ -515,7 +533,7 @@ class CursivePosStatement(Statement):
 
 class FeatureReferenceStatement(Statement):
     """Example: feature salt;"""
-    def __init__(self, location, featureName):
+    def __init__(self, featureName, location=None):
         Statement.__init__(self, location)
         self.location, self.featureName = (location, featureName)
 
@@ -527,7 +545,7 @@ class FeatureReferenceStatement(Statement):
 
 
 class IgnorePosStatement(Statement):
-    def __init__(self, location, chainContexts):
+    def __init__(self, chainContexts, location=None):
         Statement.__init__(self, location)
         self.chainContexts = chainContexts
 
@@ -556,7 +574,7 @@ class IgnorePosStatement(Statement):
 
 
 class IgnoreSubstStatement(Statement):
-    def __init__(self, location, chainContexts):
+    def __init__(self, chainContexts, location=None):
         Statement.__init__(self, location)
         self.chainContexts = chainContexts
 
@@ -584,8 +602,25 @@ class IgnoreSubstStatement(Statement):
         return "ignore sub " + ", ".join(contexts) + ";"
 
 
+class IncludeStatement(Statement):
+    def __init__(self, filename, location=None):
+        super(IncludeStatement, self).__init__(location)
+        self.filename = filename
+
+    def build(self):
+        # TODO: consider lazy-loading the including parser/lexer?
+        raise FeatureLibError(
+            "Building an include statement is not implemented yet. "
+            "Instead, use Parser(..., followIncludes=True) for building.",
+            self.location)
+
+    def asFea(self, indent=""):
+        return indent + "include(%s);" % self.filename
+
+
 class LanguageStatement(Statement):
-    def __init__(self, location, language, include_default, required):
+    def __init__(self, language, include_default=True, required=False,
+                 location=None):
         Statement.__init__(self, location)
         assert(len(language) == 4)
         self.language = language
@@ -608,7 +643,7 @@ class LanguageStatement(Statement):
 
 
 class LanguageSystemStatement(Statement):
-    def __init__(self, location, script, language):
+    def __init__(self, script, language, location=None):
         Statement.__init__(self, location)
         self.script, self.language = (script, language)
 
@@ -620,7 +655,7 @@ class LanguageSystemStatement(Statement):
 
 
 class FontRevisionStatement(Statement):
-    def __init__(self, location, revision):
+    def __init__(self, revision, location=None):
         Statement.__init__(self, location)
         self.revision = revision
 
@@ -632,7 +667,7 @@ class FontRevisionStatement(Statement):
 
 
 class LigatureCaretByIndexStatement(Statement):
-    def __init__(self, location, glyphs, carets):
+    def __init__(self, glyphs, carets, location=None):
         Statement.__init__(self, location)
         self.glyphs, self.carets = (glyphs, carets)
 
@@ -646,7 +681,7 @@ class LigatureCaretByIndexStatement(Statement):
 
 
 class LigatureCaretByPosStatement(Statement):
-    def __init__(self, location, glyphs, carets):
+    def __init__(self, glyphs, carets, location=None):
         Statement.__init__(self, location)
         self.glyphs, self.carets = (glyphs, carets)
 
@@ -660,8 +695,8 @@ class LigatureCaretByPosStatement(Statement):
 
 
 class LigatureSubstStatement(Statement):
-    def __init__(self, location, prefix, glyphs, suffix, replacement,
-                 forceChain):
+    def __init__(self, prefix, glyphs, suffix, replacement,
+                 forceChain, location=None):
         Statement.__init__(self, location)
         self.prefix, self.glyphs, self.suffix = (prefix, glyphs, suffix)
         self.replacement, self.forceChain = replacement, forceChain
@@ -691,7 +726,8 @@ class LigatureSubstStatement(Statement):
 
 
 class LookupFlagStatement(Statement):
-    def __init__(self, location, value, markAttachment, markFilteringSet):
+    def __init__(self, value=0, markAttachment=None, markFilteringSet=None,
+                 location=None):
         Statement.__init__(self, location)
         self.value = value
         self.markAttachment = markAttachment
@@ -724,7 +760,7 @@ class LookupFlagStatement(Statement):
 
 
 class LookupReferenceStatement(Statement):
-    def __init__(self, location, lookup):
+    def __init__(self, lookup, location=None):
         Statement.__init__(self, location)
         self.location, self.lookup = (location, lookup)
 
@@ -736,7 +772,7 @@ class LookupReferenceStatement(Statement):
 
 
 class MarkBasePosStatement(Statement):
-    def __init__(self, location, base, marks):
+    def __init__(self, base, marks, location=None):
         Statement.__init__(self, location)
         self.base, self.marks = base, marks
 
@@ -752,7 +788,7 @@ class MarkBasePosStatement(Statement):
 
 
 class MarkLigPosStatement(Statement):
-    def __init__(self, location, ligatures, marks):
+    def __init__(self, ligatures, marks, location=None):
         Statement.__init__(self, location)
         self.ligatures, self.marks = ligatures, marks
 
@@ -776,7 +812,7 @@ class MarkLigPosStatement(Statement):
 
 
 class MarkMarkPosStatement(Statement):
-    def __init__(self, location, baseMarks, marks):
+    def __init__(self, baseMarks, marks, location=None):
         Statement.__init__(self, location)
         self.baseMarks, self.marks = baseMarks, marks
 
@@ -792,7 +828,7 @@ class MarkMarkPosStatement(Statement):
 
 
 class MultipleSubstStatement(Statement):
-    def __init__(self, location, prefix, glyph, suffix, replacement):
+    def __init__(self, prefix, glyph, suffix, replacement, location=None):
         Statement.__init__(self, location)
         self.prefix, self.glyph, self.suffix = prefix, glyph, suffix
         self.replacement = replacement
@@ -820,8 +856,8 @@ class MultipleSubstStatement(Statement):
 
 
 class PairPosStatement(Statement):
-    def __init__(self, location, enumerated,
-                 glyphs1, valuerecord1, glyphs2, valuerecord2):
+    def __init__(self, glyphs1, valuerecord1, glyphs2, valuerecord2,
+                 enumerated=False, location=None):
         Statement.__init__(self, location)
         self.enumerated = enumerated
         self.glyphs1, self.valuerecord1 = glyphs1, valuerecord1
@@ -861,7 +897,8 @@ class PairPosStatement(Statement):
 
 
 class ReverseChainSingleSubstStatement(Statement):
-    def __init__(self, location, old_prefix, old_suffix, glyphs, replacements):
+    def __init__(self, old_prefix, old_suffix, glyphs, replacements,
+                 location=None):
         Statement.__init__(self, location)
         self.old_prefix, self.old_suffix = old_prefix, old_suffix
         self.glyphs = glyphs
@@ -892,7 +929,8 @@ class ReverseChainSingleSubstStatement(Statement):
 
 
 class SingleSubstStatement(Statement):
-    def __init__(self, location, glyphs, replace, prefix, suffix, forceChain):
+    def __init__(self, glyphs, replace, prefix, suffix, forceChain,
+                 location=None):
         Statement.__init__(self, location)
         self.prefix, self.suffix = prefix, suffix
         self.forceChain = forceChain
@@ -925,7 +963,7 @@ class SingleSubstStatement(Statement):
 
 
 class ScriptStatement(Statement):
-    def __init__(self, location, script):
+    def __init__(self, script, location=None):
         Statement.__init__(self, location)
         self.script = script
 
@@ -937,7 +975,7 @@ class ScriptStatement(Statement):
 
 
 class SinglePosStatement(Statement):
-    def __init__(self, location, pos, prefix, suffix, forceChain):
+    def __init__(self, pos, prefix, suffix, forceChain, location=None):
         Statement.__init__(self, location)
         self.pos, self.prefix, self.suffix = pos, prefix, suffix
         self.forceChain = forceChain
@@ -966,13 +1004,16 @@ class SinglePosStatement(Statement):
 
 
 class SubtableStatement(Statement):
-    def __init__(self, location):
+    def __init__(self, location=None):
         Statement.__init__(self, location)
 
 
 class ValueRecord(Expression):
-    def __init__(self, location, vertical, xPlacement, yPlacement, xAdvance, yAdvance,
-                 xPlaDevice, yPlaDevice, xAdvDevice, yAdvDevice):
+    def __init__(self, xPlacement=None, yPlacement=None,
+                 xAdvance=None, yAdvance=None,
+                 xPlaDevice=None, yPlaDevice=None,
+                 xAdvDevice=None, yAdvDevice=None,
+                 vertical=False, location=None):
         Expression.__init__(self, location)
         self.xPlacement, self.yPlacement = (xPlacement, yPlacement)
         self.xAdvance, self.yAdvance = (xAdvance, yAdvance)
@@ -1006,10 +1047,10 @@ class ValueRecord(Expression):
             vertical = self.vertical
 
         # Try format A, if possible.
-        if x == 0 and y == 0:
-            if xAdvance == 0 and vertical:
+        if x is None and y is None:
+            if xAdvance is None and vertical:
                 return str(yAdvance)
-            elif yAdvance == 0 and not vertical:
+            elif yAdvance is None and not vertical:
                 return str(xAdvance)
 
         # Try format B, if possible.
@@ -1025,7 +1066,7 @@ class ValueRecord(Expression):
 
 
 class ValueRecordDefinition(Statement):
-    def __init__(self, location, name, value):
+    def __init__(self, name, value, location=None):
         Statement.__init__(self, location)
         self.name = name
         self.value = value
@@ -1044,8 +1085,8 @@ def simplify_name_attributes(pid, eid, lid):
 
 
 class NameRecord(Statement):
-    def __init__(self, location, nameID, platformID,
-                 platEncID, langID, string):
+    def __init__(self, nameID, platformID, platEncID, langID, string,
+                 location=None):
         Statement.__init__(self, location)
         self.nameID = nameID
         self.platformID = platformID
@@ -1059,16 +1100,33 @@ class NameRecord(Statement):
             self.platEncID, self.langID, self.string)
 
     def asFea(self, indent=""):
-        plat = simplify_name_attributes(self.platformID, self.platEncID, self.langID)
+        def escape(c, escape_pattern):
+            # Also escape U+0022 QUOTATION MARK and U+005C REVERSE SOLIDUS
+            if c >= 0x20 and c <= 0x7E and c not in (0x22, 0x5C):
+                return unichr(c)
+            else:
+                return escape_pattern % c
+        encoding = getEncoding(self.platformID, self.platEncID, self.langID)
+        if encoding is None:
+            raise FeatureLibError("Unsupported encoding", self.location)
+        s = tobytes(self.string, encoding=encoding)
+        if encoding == "utf_16_be":
+            escaped_string = "".join([
+                escape(byteord(s[i]) * 256 + byteord(s[i + 1]), r"\%04x")
+                for i in range(0, len(s), 2)])
+        else:
+            escaped_string = "".join([escape(byteord(b), r"\%02x") for b in s])
+        plat = simplify_name_attributes(
+            self.platformID, self.platEncID, self.langID)
         if plat != "":
             plat += " "
-        return "nameid {} {}\"{}\";".format(self.nameID, plat, self.string)
+        return "nameid {} {}\"{}\";".format(self.nameID, plat, escaped_string)
 
 
 class FeatureNameStatement(NameRecord):
     def build(self, builder):
         NameRecord.build(self, builder)
-        builder.add_featureName(self.location, self.nameID)
+        builder.add_featureName(self.nameID)
 
     def asFea(self, indent=""):
         if self.nameID == "size":
@@ -1082,8 +1140,8 @@ class FeatureNameStatement(NameRecord):
 
 
 class SizeParameters(Statement):
-    def __init__(self, location, DesignSize, SubfamilyID, RangeStart,
-                 RangeEnd):
+    def __init__(self, DesignSize, SubfamilyID, RangeStart, RangeEnd,
+                 location=None):
         Statement.__init__(self, location)
         self.DesignSize = DesignSize
         self.SubfamilyID = SubfamilyID
@@ -1101,8 +1159,50 @@ class SizeParameters(Statement):
         return res + ";"
 
 
+class CVParametersNameStatement(NameRecord):
+    def __init__(self, nameID, platformID, platEncID, langID, string,
+                 block_name, location=None):
+        NameRecord.__init__(self, nameID, platformID, platEncID, langID,
+                            string, location=location)
+        self.block_name = block_name
+
+    def build(self, builder):
+        item = ""
+        if self.block_name == "ParamUILabelNameID":
+            item = "_{}".format(builder.cv_num_named_params_.get(self.nameID, 0))
+        builder.add_cv_parameter(self.nameID)
+        self.nameID = (self.nameID, self.block_name + item)
+        NameRecord.build(self, builder)
+
+    def asFea(self, indent=""):
+        plat = simplify_name_attributes(self.platformID, self.platEncID,
+                                        self.langID)
+        if plat != "":
+            plat += " "
+        return "name {}\"{}\";".format(plat, self.string)
+
+
+class CharacterStatement(Statement):
+    """
+    Statement used in cvParameters blocks of Character Variant features (cvXX).
+    The Unicode value may be written with either decimal or hexadecimal
+    notation. The value must be preceded by '0x' if it is a hexadecimal value.
+    The largest Unicode value allowed is 0xFFFFFF.
+    """
+    def __init__(self, character, tag, location=None):
+        Statement.__init__(self, location)
+        self.character = character
+        self.tag = tag
+
+    def build(self, builder):
+        builder.add_cv_character(self.character, self.tag)
+
+    def asFea(self, indent=""):
+        return "Character {:#x};".format(self.character)
+
+
 class BaseAxis(Statement):
-    def __init__(self, location, bases, scripts, vertical):
+    def __init__(self, bases, scripts, vertical, location=None):
         Statement.__init__(self, location)
         self.bases = bases
         self.scripts = scripts
@@ -1119,7 +1219,7 @@ class BaseAxis(Statement):
 
 
 class OS2Field(Statement):
-    def __init__(self, location, key, value):
+    def __init__(self, key, value, location=None):
         Statement.__init__(self, location)
         self.key = key
         self.value = value
@@ -1144,7 +1244,7 @@ class OS2Field(Statement):
 
 
 class HheaField(Statement):
-    def __init__(self, location, key, value):
+    def __init__(self, key, value, location=None):
         Statement.__init__(self, location)
         self.key = key
         self.value = value
@@ -1159,7 +1259,7 @@ class HheaField(Statement):
 
 
 class VheaField(Statement):
-    def __init__(self, location, key, value):
+    def __init__(self, key, value, location=None):
         Statement.__init__(self, location)
         self.key = key
         self.value = value
