@@ -2,13 +2,34 @@
 from __future__ import print_function, division, absolute_import
 from fontTools.misc.py23 import *
 
-__all__ = ['normalizeLocation', 'supportScalar', 'VariationModel']
+__all__ = ['normalizeValue', 'normalizeLocation', 'supportScalar', 'VariationModel']
+
+
+def normalizeValue(v, triple):
+	"""Normalizes value based on a min/default/max triple.
+	>>> normalizeValue(400, (100, 400, 900))
+	0.0
+	>>> normalizeValue(100, (100, 400, 900))
+	-1.0
+	>>> normalizeValue(650, (100, 400, 900))
+	0.5
+	"""
+	lower, default, upper = triple
+	assert lower <= default <= upper, "invalid axis values: %3.3f, %3.3f %3.3f"%(lower, default, upper)
+	v = max(min(v, upper), lower)
+	if v == default:
+		v = 0.
+	elif v < default:
+		v = (v - default) / (default - lower)
+	else:
+		v = (v - default) / (upper - default)
+	return v
 
 def normalizeLocation(location, axes):
 	"""Normalizes location based on axis min/default/max values from axes.
 	>>> axes = {"wght": (100, 400, 900)}
 	>>> normalizeLocation({"wght": 400}, axes)
-	{'wght': 0}
+	{'wght': 0.0}
 	>>> normalizeLocation({"wght": 100}, axes)
 	{'wght': -1.0}
 	>>> normalizeLocation({"wght": 900}, axes)
@@ -21,9 +42,9 @@ def normalizeLocation(location, axes):
 	{'wght': -1.0}
 	>>> axes = {"wght": (0, 0, 1000)}
 	>>> normalizeLocation({"wght": 0}, axes)
-	{'wght': 0}
+	{'wght': 0.0}
 	>>> normalizeLocation({"wght": -1}, axes)
-	{'wght': 0}
+	{'wght': 0.0}
 	>>> normalizeLocation({"wght": 1000}, axes)
 	{'wght': 1.0}
 	>>> normalizeLocation({"wght": 500}, axes)
@@ -38,27 +59,21 @@ def normalizeLocation(location, axes):
 	>>> normalizeLocation({"wght": 500}, axes)
 	{'wght': -0.5}
 	>>> normalizeLocation({"wght": 1000}, axes)
-	{'wght': 0}
+	{'wght': 0.0}
 	>>> normalizeLocation({"wght": 1001}, axes)
-	{'wght': 0}
+	{'wght': 0.0}
 	"""
 	out = {}
-	for tag,(lower,default,upper) in axes.items():
-		assert lower <= default <= upper, "invalid axis values"
-		v = location.get(tag, default)
-		v = max(min(v, upper), lower)
-		if v == default:
-			v = 0
-		elif v < default:
-			v = (v - default) / (default - lower)
-		else:
-			v = (v - default) / (upper - default)
-		out[tag] = v
+	for tag,triple in axes.items():
+		v = location.get(tag, triple[1])
+		out[tag] = normalizeValue(v, triple)
 	return out
 
-def supportScalar(location, support):
+def supportScalar(location, support, ot=True):
 	"""Returns the scalar multiplier at location, for a master
-	with support.
+	with support.  If ot is True, then a peak value of zero
+	for support of an axis means "axis does not participate".  That
+	is how OpenType Variation Font technology works.
 	>>> supportScalar({}, {})
 	1.0
 	>>> supportScalar({'wght':.2}, {})
@@ -67,13 +82,29 @@ def supportScalar(location, support):
 	0.1
 	>>> supportScalar({'wght':2.5}, {'wght':(0,2,4)})
 	0.75
+	>>> supportScalar({'wght':2.5, 'wdth':0}, {'wght':(0,2,4), 'wdth':(-1,0,+1)})
+	0.75
+	>>> supportScalar({'wght':2.5, 'wdth':.5}, {'wght':(0,2,4), 'wdth':(-1,0,+1)}, ot=False)
+	0.375
+	>>> supportScalar({'wght':2.5, 'wdth':0}, {'wght':(0,2,4), 'wdth':(-1,0,+1)})
+	0.75
+	>>> supportScalar({'wght':2.5, 'wdth':.5}, {'wght':(0,2,4), 'wdth':(-1,0,+1)})
+	0.75
 	"""
 	scalar = 1.
 	for axis,(lower,peak,upper) in support.items():
-		if axis not in location:
-			scalar = 0.
-			break
-		v = location[axis]
+		if ot:
+			# OpenType-specific case handling
+			if peak == 0.:
+				continue
+			if lower > peak or peak > upper:
+				continue
+			if lower < 0. and upper > 0.:
+				continue
+			v = location.get(axis, 0.)
+		else:
+			assert axis in location
+			v = location[axis]
 		if v == peak:
 			continue
 		if v <= lower or upper <= v:
@@ -122,12 +153,12 @@ class VariationModel(object):
 	 {0: 1.0},
 	 {0: 1.0},
 	 {0: 1.0, 4: 1.0, 5: 1.0},
-	 {0: 1.0, 3: 0.75, 4: 0.25, 5: 1.0, 6: 0.25},
+	 {0: 1.0, 3: 0.75, 4: 0.25, 5: 1.0, 6: 0.6666666666666666},
 	 {0: 1.0,
 	  3: 0.75,
 	  4: 0.25,
 	  5: 0.6666666666666667,
-	  6: 0.16666666666666669,
+	  6: 0.4444444444444445,
 	  7: 0.6666666666666667}]
 	"""
 
@@ -140,7 +171,7 @@ class VariationModel(object):
 		self.mapping = [self.locations.index(l) for l in locations] # Mapping from user's master order to our master order
 		self.reverseMapping = [locations.index(l) for l in self.locations] # Reverse of above
 
-		self._computeMasterSupports(axisPoints)
+		self._computeMasterSupports(axisPoints, axisOrder)
 
 	@staticmethod
 	def getMasterLocationsSortKeyFunc(locations, axisOrder=[]):
@@ -152,7 +183,7 @@ class VariationModel(object):
 			axis = next(iter(loc))
 			value = loc[axis]
 			if axis not in axisPoints:
-				axisPoints[axis] = {0}
+				axisPoints[axis] = {0.}
 			assert value not in axisPoints[axis]
 			axisPoints[axis].add(value)
 
@@ -191,7 +222,7 @@ class VariationModel(object):
 		else:
 			return value
 
-	def _computeMasterSupports(self, axisPoints):
+	def _computeMasterSupports(self, axisPoints, axisOrder):
 		supports = []
 		deltaWeights = []
 		locations = self.locations
@@ -203,7 +234,10 @@ class VariationModel(object):
 				if not axis in loc:
 					continue
 				locV = loc[axis]
-				box[axis] = (self.lowerBound(locV, values), locV, self.upperBound(locV, values))
+				if locV > 0:
+					box[axis] = (0, locV, max({locV}|values))
+				else:
+					box[axis] = (min({locV}|values), locV, 0)
 
 			locAxes = set(loc.keys())
 			# Walk over previous masters now
@@ -214,20 +248,38 @@ class VariationModel(object):
 				# If it's NOT in the current box, it does not participate
 				relevant = True
 				for axis, (lower,_,upper) in box.items():
-					if axis in m and not (lower < m[axis] < upper):
+					if axis not in m or not (lower < m[axis] < upper):
 						relevant = False
 						break
 				if not relevant:
 					continue
-				# Split the box for new master
-				for axis,val in m.items():
+
+				# Split the box for new master; split in whatever direction
+				# that has largest range ratio.  See commit for details.
+				orderedAxes = [axis for axis in axisOrder if axis in m.keys()]
+				orderedAxes.extend([axis for axis in sorted(m.keys()) if axis not in axisOrder])
+				bestAxis = None
+				bestPercentage = -1
+				for axis in orderedAxes:
+					val = m[axis]
 					assert axis in box
 					lower,locV,upper = box[axis]
+					newLower, newUpper = lower, upper
 					if val < locV:
-						lower = val
+						newLower = val
+						percentage = (val - locV) / (lower - locV)
 					elif locV < val:
-						upper = val
-					box[axis] = (lower,locV,upper)
+						newUpper = val
+						percentage = (val - locV) / (upper - locV)
+					if percentage > bestPercentage:
+						bestPercentage = percentage
+						bestAxis = axis
+						bestLower = newLower
+						bestUpper = newUpper
+						bestLocV = locV
+
+				if bestAxis:
+					box[bestAxis] = (bestLower,bestLocV,bestUpper)
 			supports.append(box)
 
 			deltaWeight = {}
@@ -242,36 +294,72 @@ class VariationModel(object):
 		self.deltaWeights = deltaWeights
 
 	def getDeltas(self, masterValues):
-		count = len(self.locations)
 		assert len(masterValues) == len(self.deltaWeights)
 		mapping = self.reverseMapping
 		out = []
 		for i,weights in enumerate(self.deltaWeights):
 			delta = masterValues[mapping[i]]
 			for j,weight in weights.items():
-				delta -= out[j] * weight
+				try:
+					delta -= out[j] * weight
+				except OverflowError:
+					# if it doesn't fit signed shorts, retry with doubles
+					delta._ensureFloat()
+					delta -= out[j] * weight
 			out.append(delta)
 		return out
 
-	def interpolateFromDeltas(self, loc, deltas):
+	def getScalars(self, loc):
+		return [supportScalar(loc, support) for support in self.supports]
+
+	@staticmethod
+	def interpolateFromDeltasAndScalars(deltas, scalars):
 		v = None
-		supports = self.supports
-		assert len(deltas) == len(supports)
-		for i,(delta,support) in enumerate(zip(deltas, supports)):
-			scalar = supportScalar(loc, support)
+		assert len(deltas) == len(scalars)
+		for i,(delta,scalar) in enumerate(zip(deltas, scalars)):
 			if not scalar: continue
 			contribution = delta * scalar
-			if i == 0:
+			if v is None:
 				v = contribution
 			else:
 				v += contribution
 		return v
 
+	def interpolateFromDeltas(self, loc, deltas):
+		scalars = self.getScalars(loc)
+		return self.interpolateFromDeltasAndScalars(deltas, scalars)
+
 	def interpolateFromMasters(self, loc, masterValues):
 		deltas = self.getDeltas(masterValues)
 		return self.interpolateFromDeltas(loc, deltas)
 
+	def interpolateFromMastersAndScalars(self, masterValues, scalars):
+		deltas = self.getDeltas(masterValues)
+		return self.interpolateFromDeltasAndScalars(deltas, scalars)
+
+
+def main(args):
+	from fontTools import configLogger
+
+	args = args[1:]
+
+	# TODO: allow user to configure logging via command-line options
+	configLogger(level="INFO")
+
+	axes = [chr(c) for c in range(ord('A'), ord('Z')+1)]
+
+	locs = [dict(zip(axes, (float(v) for v in s.split(',')))) for s in args]
+	model = VariationModel(locs)
+	from pprint import pprint
+	print("Sorted locations:")
+	pprint(model.locations)
+	print("Supports:")
+	pprint(model.supports)
 
 if __name__ == "__main__":
 	import doctest, sys
+
+	if len(sys.argv) > 1:
+		sys.exit(main(sys.argv))
+
 	sys.exit(doctest.testmod().failed)
